//Ethan Alexander Shulman "Obscurity" December 7, 2021
#version 150
precision highp float;

//UNIFORMS/ATTRIBUTES
in vec3 in_Position;
out float time;
out vec4 uv;

uniform sampler2D Buffer0;

layout(std140) uniform ShaderviewBlock {
	vec4 mouse;
	float screenX,screenY;
	int frame,framerate;
};



//CONFIG/CONSTANTS

//Rendering config
//#define DEBUG 1//Debug rendering displays normals. Comment out to disable.
#define STEPS 200//Total path marching steps.
#define EPS 0.005//Marching epsilon.
#define STEP_DITHER 1.//Dither marching step size, hides banding via blurring. Comment out to disable dithering.
#define BACKGROUND_DISTANCE 2.//Distance from geometry before background emission takes effect.
//#define COLOR_MAX 1.//Decreases noise caused by chromatic separation at the cost of accuracy. Comment out to disable.

//Animation config
//#define ANIMATION 1//Toggle animation mode.
#define FSAMPLES 100//Paths sampled per animation frame.


//CONSTANTS
#define PI 3.14159265


//GLOBAL VARIABLES
vec4 baseRand;//random hash from screen uv and frame
float nsampling;//flag = 1 if sampling normal, used for cancelling volume dither




//UTILITY FUNCTIONS
float smin(float a, float b, float fade, float mag) {
	float al = clamp(a*fade,0.,1.),
		bl = clamp(b*fade,0.,1.);
	return min(a,b)-mag*max(0.,1.-al-bl);
}

vec3 hue(float h) {
	return clamp(abs(mod(h*6.+vec3(0.,4.,3.),6.)-3.)-1.,0.,1.);
}

float twave(float x) {
	return 1.-abs(fract(x)*2.-1.);
}

//2d rotation matrix
mat2 r2(float a) {
	float s = sin(a), c = cos(a);
	return mat2(c,s,-s,c);
}

//tangent and bitangent from normal
mat2x3 tangentBitangent(vec3 norm) {
	mat2x3 t;
	if (abs(norm.x) > abs(norm.y)) t[0] = normalize(vec3(-norm.z,0,norm.x));
	else t[0] = normalize(vec3(0,norm.z,-norm.y)); 
	t[1] = cross(norm,t[0]);
	return t;
}

//uvec4 seed, returns uvec4 hash
uvec4 hash(uvec4 seed) {
    uvec4 h = (0x6A7F8FAAu^seed)*0x01000193u;
    h = ((h.wxyz>>3u)^h^seed.yzwx)*0x01000193u;
    h = ((h.zwxy>>8u)^h^seed.wxyz)*0x01000193u;
    return h^(h>>11u);
}
#define I2F (1./float(0xFFFFFFFFu))


//distance dithering for volumetrics
#define START_VOLUME(posSeed) vec2 opacityRand = fract(baseRand.zw+128.*abs(dot(posSeed,baseRand.zxy))*baseRand.xy)
#define VOLUME(dst,skip,alpha,isotropy) (max(dst,-1e3+(skip*(1.+opacityRand.y)+1e3)*step(alpha+abs(nsampling),opacityRand.x))+max(0.,nsampling)*(isotropy)*opacityRand.y)

//dither materials for chromatic separation
#define START_CHROMA(posSeed) float chromaFrame = fract(baseRand.x+156.*dot(abs(posSeed),baseRand.zwy))
#define CHROMA(mat,rc,gc,bc) if (chromaFrame < 1.) {mat[0]*=vec3(3,0,0);rc;}\
else if (chromaFrame < 2.) {mat[0]*=vec3(0,3,0);gc;}\
else {mat[0]*=vec3(0,0,3);bc;}




//PATH TRACING FUNCTIONS

//defines scene geometry, returns distance(x) and material id(y)
vec2 geometry(vec3 p);

/*defines scene materials
mat[0] = color
mat[1].x = roughness, mat[1].y = emission*/
mat2x3 material(vec3 p, float matId);

//returns normal at point
vec3 normal(vec3 p) {
	#define NEPS vec3(EPS*.5,0,0)
	nsampling = 1.;
	vec3 n = normalize(vec3(geometry(p+NEPS).x,geometry(p+NEPS.yxz).x,geometry(p+NEPS.yzx).x)-geometry(p).x);
	nsampling = 0.;
	return n!=n?vec3(0,1,0):n;
}

//defines scene background, returns background color at direction and position
vec3 background(vec3 rp, vec3 rd);

//pathtrace from camera perspective
vec4 pathtrace(vec3 cameraPos, vec3 cameraRot, float camNear, float fovTangent) {
	baseRand = hash(uvec4(uv.yzw,frame))*I2F;
	
	vec3 rp = cameraPos,//ray position
		rd = normalize(vec3(uv.xy+(baseRand.xy*2.-1.)/screenY,fovTangent)),//ray direction
		c = vec3(1),//color tint
		l = vec3(0);//absored light emission

	if (cameraRot.z != 0.) {
		mat2 rm = r2(cameraRot.z);
		rd.xy *= rm;
		//rp.xy *= rm;//roll
	}
	if (cameraRot.y != 0.) {
		mat2 rm = r2(cameraRot.y);
		rd.yz *= rm;
		//rp.yz *= rm;//pitch
	}
	if (cameraRot.x != 0.) {
		mat2 rm = r2(cameraRot.x);
		rd.xz *= rm;
		//rp.xz *= rm;//yaw
	}
	rp += rd*camNear;

	//main path tracing loop
	for (int i = 0; i < STEPS; i++) {
		vec2 d = geometry(rp);
		if (d.x <= EPS) {
			//back up ray
			rp -= rd*clamp(EPS*2.-d.x,-EPS*4.,EPS*4.);
			#ifdef DEBUG
			return vec4(normal(rp)*.5+.5,1);
			#endif
			
			//material absorption
			mat2x3 m = material(rp,d.y);
			c *= m[0];
			#ifdef COLOR_MAX
			vec3 normc = c;
			float clen = length(normc);
			if (clen > 1.7320508075688772*COLOR_MAX) normc *= (1.7320508075688772*COLOR_MAX)/clen;
			l += normc*m[1].y;
			#else
			l += c*m[1].y;
			#endif

			//random microsurface bump based on roughness
			baseRand.xy = fract(baseRand.xy+baseRand.zw+abs(rp.xy-rp.z));
			float ct = pow(baseRand.x,m[1].x), st = sqrt(1.-ct*ct),
				pt = baseRand.y*PI*2.;
			vec3 nrm = normal(rp);
			mat2x3 tb = tangentBitangent(nrm);
			vec3 bump = normalize(nrm*ct + tb[0]*(st*cos(pt)) + tb[1]*(st*sin(pt)));
			
			//reflect ray
			rd = reflect(rd,bump);
		} else {
			#ifdef STEP_DITHER
			rp += rd*max(EPS,d.x-baseRand.z*EPS*STEP_DITHER);
			#else
			rp += rd*d.x;
			#endif
		}
	}
	
	//add background emission
	nsampling = -1.;
	l += c*background(rp,rd)*clamp(geometry(rp).x/BACKGROUND_DISTANCE,0.,1.);
	
	return vec4(l,1);
}






//SCENE FUNCTIONS
vec2 geometry(vec3 p) {
	START_VOLUME(p);
	
	//water
	float d = p.y, mat = 0.;
	if (p.y < 4.) {//deform only within bounds for optimization
		for (float f = 1.; f < 10.; f++) {
			float a = f*.3, s = .5+f*f*.1;
			d -= twave(s*abs(dot(p.xz,vec2(cos(a),sin(a)))))*.05/s;
		}
	}
	float wd = d;
	d = VOLUME(d, .1, .05, 0.01);
	
	//floating objects
	if (wd < 0.) {
		float bd = length(max(abs(mod(abs(p-vec3(0,.7,0)),2.8)-1.4)-vec3(.1,.17,.08),0.))-.1;
		if (bd < d) {
			d = bd;
			mat = 2.;
		}
	}
	
	//clouds
	if (wd > 0.2) {
		float cd = max(p.y-3., abs(mod(abs(wd),.6)-.3)-.05);
		cd = VOLUME(cd, 1., .1, 0.);
		if (cd < d) {
			d = cd;
			mat = 1.;
		}
	}
	
	return vec2(d,mat);
}

mat2x3 material(vec3 p, float matId) {
	if (matId == 2.) {
		//floater material
		return mat2x3(vec3(0.), vec3(.2,0.,0.));
	}
	if (matId == 1.) {
		//cloud material
		return mat2x3(vec3(.5), vec3(.4,0.,0.));
	}
	//water material
	return mat2x3(mix(vec3(.001,.04,.075),vec3(.2,.27,.6),clamp((p.y-.01)*5.,0.,1.)), vec3(0.00001,0.,0));
}

vec3 background(vec3 rp, vec3 rd) {
	return vec3(5,7,9)*max(0.,dot(rd,normalize(vec3(1,1,1)))-.8)*10.;
}



//Vertex and pixel shader
void vert() {
	#ifdef ANIMATION
	time = mod(float(frame)/float(framerate*FSAMPLES),10.);
	#else
	time = 0.;
	#endif
	//uv.xy = aspect correct normalized coordinates, uv.zw = screen pixel coordinates
	uv = vec4(in_Position.xy*vec2(screenX/screenY,1),(in_Position.xy*.5+.5)*vec2(screenX,screenY));
	gl_Position = vec4(in_Position.xy,0,1);
}
void frag() {
	//progressive rendering
	float mt = (time-4.)/6.;
	vec4 l = pathtrace(vec3(10.,3.,20),//cam position
						vec3(.6,-0.8,0),//cam rotation
							 0.,//camera near
							 1.3);//field of view
	#ifdef ANIMATION
	gl_FragColor = (frame%FSAMPLES==0?vec4(0):texelFetch(Buffer0,ivec2(uv.zw),0))+l*1e-2;
	#else
	gl_FragColor = (frame==0?vec4(0):texelFetch(Buffer0,ivec2(uv.zw),0))+l*1e-2;
	#endif
}
